<?php
#
# This class implements a compiler for a simple language that describes
# rewrites that have to be done on Phorum data.
#
# Usage:
# ======
#
# $compiler = new rewrite_rules_compiler();
# $code = $compiler->compile($source_file);
#
# Command syntax:
# ===============
#
#   Restrict code to run only on a certain page
#   -------------------------------------------
#
#   page <page name>
#     ...
#   end page
#
#   Process all records in an array
#   -------------------------------
#   Replace commands that are placed within an array block will be
#   applied to all records in the specified array.
#   
#   array <variable>
#     ...
#   end array
# 
#   Process a single record
#   -----------------------
#
#   record <variable>
#     ...
#   end record
#
#   Do a rewrite on a record
#   ------------------------
#
#   rewrite <target variable> with <function>(<argument variable>)
#
#   Functions that are available in the real_name mod are:
#
#   - displayname  : the display name for a user, based on user_id
#   - profile_url  : the profile url for a user, based on user_id
#   - profile_link : a full clickable link to the user profile based on user_id
# 
# Variable syntax:
# ================
#
#   [variable]        $GLOBALS["PHORUM"]["variable"]
#   [var1->var2]      $GLOBALS["PHORUM"]["var1"]["var2"]
#   variable          $current_record["variable"]
#
# Example:
# ========
#
#   page some_page
#     array [DATA->LIST]
#       rewrite the_name with displayname(the_id)
#       rewrite the_linkedname with profile_link(the_id)
#     end array
#     record [DATA->MYINFO]
#       rewrite name with displayname([user->user_id])
#     end record
#   end page
#
#   This code will only run on the phorum_page "some_page". It will first
#   walk through the array $PHORUM["DATA"]["LIST"]. To each record in that
#   array, two rewrites are applied. The field "the_name" will be replaced
#   by the display name, using the "the_id" field in the same record as the
#   user_id to use as the imput for the displayname function. The field
#   "the_linkname" will be rewritten in the same way, only using the 
#   rewrite function "profile_link".
#   After this, the record $PHORUM["DATA"]["MYINFO"] is rewritten. The
#   field "name" is rewritten with the display name, using
#   $PHORUM["user"]["user_id"] as the user_id argument for the displayname
#   function.
#
class rewrite_rules_compiler
{
    var $stack  = array();

    function compile_variable($var)
    {
        $origvar = $var;
        $var = str_replace("->", "']['", $var);

        if (preg_match('/^\[(.+)\]$/', $var, $m)) {
            return "\$GLOBALS['PHORUM']['$m[1]']";
        } else {
            if (count($this->stack) == 0) {
                die ("Error in rewrite input: no stack, " .
                     "but relative variable " . htmlspecialchars($origvar) .
                     " used.");
            }

            $stack_last = $this->stack[count($this->stack)-1];
            $idx = is_null($stack_last[1]) ? '' : '[' . $stack_last[1] . ']';
            return "{$stack_last[0]}{$idx}['$var']";
        }
    }

    function dynamic_variable($prefix)
    {
        static $var_i = 0;
        $var_i ++;
        return '$' . $prefix . '_' . $var_i;
    }

    function compile($input, $code = array())
    {
        $linenr = 0;
        $indent = 0;
        $this->stack = array();
        $page   = '';
        $chunk  = '';

        if (! is_array($code)) {
            die ("Illegal \$code argument for the real_name rewrite rules " .
                 "compile() function (it has to be an array)");
        }

        $fp = fopen($input, "r");
        if (! $fp) die ("Cannot open phorum rewrite rules file " .
                        htmlspecialchars($input) . " for reading");

        if (! isset($code[$page])) $code[$page] = ''; 

        while ($line = fgets($fp))
        {
            $linenr++;

            // Skip empty lines and comments.
            if (preg_match('/^(\s*|\s*#.*)$/', $line)) continue;

            // Strip all whitespace and newline stuff.
            $line = preg_replace('/^\s+|\r|\n|\s+$/', '', $line);

            // End of some block? Then we need to lower the indent here
            // for getting the automatic indention right.
            if (substr($line, 0, 4) == 'end ') {
                $indent--;
            }

            // Add automatic indention.
            $ind =  str_repeat(" ", $indent*2);
            $chunk .= "{$ind}### $line\n";
            $chunk .= $ind;

            // page <page name>
            if (preg_match('/^page\s+(\S+)$/', $line, $m)) {
                $page = $m[1];
                if (! isset($code[$page])) $code[$page] = ''; 
                $indent++;
            }

            // end page
            elseif ($line == 'end page') {
                $page = '';
            }

            // array <array name>
            elseif (preg_match('/^array\s+(\S+)$/', $line, $m)) {
                $var = $this->compile_variable($m[1]);
                $idvar = $this->dynamic_variable('id');
                $this->stack[] = array($var, $idvar);
                $indent++;

                $code[$page] .= "if (isset($var) && is_array($var)) {\n" .
                                "$ind foreach ($var as $idvar => \$tmp) {\n";
            }

            // end array
            elseif ($line == 'end array') {
                array_pop($this->stack);
                $code[$page] .= " }\n$ind}\n";
            }

            // record <record name>
            elseif (preg_match('/^record\s+(\S+)$/', $line, $m)) {
                $var = $this->compile_variable($m[1]);
                $this->stack[] = array($var, NULL);
                $indent++;

                $code[$page] .= "if (isset($var) && is_array($var)) {\n";
            }

            // end record
            elseif ($line == 'end record') {
                array_pop($this->stack);
                $code[$page] .= "}\n";
            }

            // rewrite <field> with <function>(<argument>)
            elseif (preg_match('/^rewrite\s+(\S+)\s+with\s+([^\(]+)\(([^\)]+)\)$/', $line, $m)) {
                $targetvar = $this->compile_variable($m[1]);
                $call = 'mod_real_name_rewrite';
                $functioncmd = "'" . addslashes($m[2]) . "'";
                $functionarg = $this->compile_variable($m[3]);

                $code[$page] .= "if (isset($targetvar) && isset($functionarg)) {\n" .
                                "$ind  \$rewrite = $call($functioncmd, $functionarg);\n".
                                "$ind  if (\$rewrite !== NULL) $targetvar = \$rewrite;\n".
                                "$ind}\n";
            }

            // Syntax error
            else {
                die ("Syntax error on line $linenr of rewrite file " . 
                     htmlspecialchars($input));
            }
        }

        return $code;
    }
}
?>
